{-
Time-stamp: <Wed Sep 29 2010 20:56:38 Stardate: Stardate: [-28]3758.94 hwloidl>
$Id: LinSolv.hs,v 1.5.2.5 2001/09/29 02:19:37 hwloidl Exp $

Module for computing the solution of a system of linear system of equations
in several variables.
The rhs of the system of equations is the first parameter (given as a 
square matrix). The lhs of the system is the second parameter (given as a
vector).

Note: This version uses Strategies to specify parallelism. It defines the 
      Strategies on top-level i.e. in the body of the let in the main 
      linsolv function. Of the various Strategies used, Strategies 26,
      27 and 28 correspond to those discussed in my LinSolv paper:
        http://www.cee.hw.ac.uk/~dsg/gph/papers/ps/linsolv.ps.gz
---------------------------------------------------------------------------- -}

-- @node LinSolv top level function, , ADT Matrix, Top
-- @chapter LinSolv top level function

module LinSolv(linSolv) where

#warning "This version is deprecated; use LinSolvS.hs instead"

-- version = "LinSolv: $Revision: 1.5.2.5 $"

-- @node Imports, Auxiliary functions
-- @section Imports

-- @include ModArithm.hs
import ModArithm {- (modHom, Hom(hom))-}

-- @include Matrix-list.hs
import Matrix (SqMatrix, Vector, vector, matBounds, unSqMatrix,
#ifndef LU_DET
               determinant, 
#endif
               transp, replaceColumn, size,
               maxElem, maxElemVec, scalarMult, vecScalarQuot,
               matGcd, vecGcd, matHom, vecHom) 

#ifdef LU_DET
import qualified Matrix (determinant)
#endif

-- @include CRA.hs
import CRA (seq_list_CRA)

-- @include LUDecomp.hs
import LU (det)  -- det computation via LU decomposition

-- @include Ratio.hs
import Ratio

#ifndef __EDEN__
import Strategies
#endif

#ifdef __EDEN__
import Eden hiding(parMap, farm)
import EdenSkel
#endif

#ifdef TRACE
--import Unsafe(trace)
import GlaExts(trace)
#endif

#ifdef LU_DET
slow_determinant = Matrix.determinant

-- determinant via LU Decomposition
determinant :: SqMatrix Integer -> Integer
determinant m = let 
                  funny_array = (fmap toRational (unSqMatrix m))
                  d = det funny_array 
                in 
# ifdef TRACE
	        trace ("Passing a funny array tp det: " ++ (show funny_array)) $
# endif
                if denominator d /= 1
                  then error ("det: denom /= 1" ++ show d)
                  else numerator d
#endif

-- ---------------------------------------------------------------------------
-- @node Auxiliary functions, Main Fct, Imports
-- @section Auxiliary functions
-- ---------------------------------------------------------------------------

primes :: [Integer]
primes = 2 : oddPrimesFrom 3
         where oddPrimesFrom n | isPrime n primes = n : oddPrimesFrom (n+2)
                               | otherwise        =     oddPrimesFrom (n+2)

isPrime :: Integer -> [Integer] -> Bool
isPrime n (l1:ls) | n < l1*l1        = True
                  | n `mod` l1 == 0  = False
                  | otherwise        = isPrime n ls

projection :: Integer -> [[Integer]] -> [Integer]
projection i = _scc_ "projection" ( map (!! (fromInteger i)) )
--projection i [] = [] 
--projection i (ls1:lss) = ls1!!(i-1) : projection i lss
 
fact :: Integer -> Integer
fact 0   = 1
fact (n+1) = (n+1) * fact n

-- ---------------------------------------------------------------------------
-- @node Main Fct, Junk, Auxiliary functions
-- @section Main Fct
--
-- Function for solving a system of linear equations
-- ---------------------------------------------------------------------------

-- @cindex linSolv
linSolv :: SqMatrix Integer ->                -- nxn matrix A
           Vector Integer ->                  -- n vector b
           (Vector Integer, Integer, Integer) -- n vector x s.t. A*x=b
linSolv a b = 
  let 
   {- Step1: -}
    n = toInteger (size a) 
    s = max (maxElem a) (maxElemVec b)
    pBound :: Integer
    pBound = 2 * s^n * fact n 

   {- Step2: Computation of solutions in Z/p -}
    guessedNoOfPrimes :: Integer
    guessedNoOfPrimes =  count primes' 

    xList,xList_unlucky :: [[Integer]]
    xList = filter lucky 
	    -- map (t )	
		xList_all
	   where t x | lucky x = x
		     | otherwise = 1:0:error "no sol. vector"

    xList_unlucky = filter (not.lucky) xList_all

    lucky (_:0:_) = False
    lucky _ = True

    -- needed to add more tasks, for the dropped of primes:
    unluckyPrimes = projection 0 xList_unlucky

   {- We need a way to control the speculation in Eden.
      The number of homomorphic images needed is determined by
      the product of the initial lucky prime numbers which can be
      approximated by the product of all initial prime numbers.
   -}

# ifdef UNLIMITED
    primes' = 
# ifdef TRACE
	      map (\x -> trace (show x) x)
# endif
		primes

# else	
   {- split primes in first part needed anyway
	and the rest that is included only on demand -}
    primes', p_needed, p_speculative :: [Integer]
    (p_needed, p_speculative) = splitAt ( 1 + fromIntegral guessedNoOfPrimes) primes
    primes' = p_needed ++ (additional (product p_needed) 
			     unluckyPrimes p_speculative)


   {- add another prime number for each unlucky prime.
      This results in more tasks, as if only the product
	is kept over pBound. But the latter had a termination problem. 
								(UK)
   -}
    additional :: Integer -> [Integer] -> [Integer] -> [Integer]
    additional pMax (pu:pus) (pn:pns)
	| -- pMax' <= pBound = (
	  True = (
#ifdef TRACE
	   	   trace (show pn) 
#endif
		    pn) 
		    : (additional (pMax --'*pn -- currently not used, so do 
                                               -- create unnecessary thunks!
					) pus pns)
        | otherwise      = additional pMax' pus (pn:pns)
	where pMax' = pMax `div` pu
# endif

#ifdef __EDEN__
    np :: Int
# ifdef MINUS1PE
    np | noPe == 1 = 1
       | otherwise = noPe -1
# else
    np = noPe
# endif

    xList_all :: [[Integer]]  -- inf list of solutions in hom images
#if (SKEL==3)
	-- WORKPOOL
# ifndef PREFETCH
#  define PREFETCH 2
# endif
    xList_all = (workpool PREFETCH np 1 get_homSol primes')
#elif (SKEL==4)
	{- PARMAP
           with the restriction in the primes the parMap should work,	
	   as only a finite number of processes are created,
	   whereas the list of processes is expanded lately. -}
    xList_all = parMap get_homSol primes'
#elif (SKEL==3)
	-- DM
# ifndef PREFETCH
#  defined PREFETCH 2
# endif

    {- For the direct mapping version, a iterative approach is used: 
	First for the starting primes the solutions are
	evaluated. 
	If this is not enother, another chunk is to be computed.
	And so on.
    -}
    xList_all = concat xLists_all
    xLists_all = (dm [solveMod i | i <- [0..np-1]] (repeat steps))
       

    steps = error "TODO: finish definition"
          where
	    f xss pAux 
		| (pAux * pPart >= pBound) = []
		| otherwise = (np*PREFETCH):
			      (f (tail (dropWhile (not.null) xss)) 
						(pAux * pPart))
		where pPart = produce (takeWhile (not.null) xss)
		      produce xs = ...

    solveMod :: Int -> Process [Int] [[Integer]]
    solveMod i = process steps ->
 		       (map gen_xList part_primes)
    where 
      part_primes = takeEach np 
			(drop i (primesChunckwise))
      primesChunkwise = f steps primes
      f [] ps = []
      f (s:ss) ps = start ++ [] ++ f ss rest
		  where (start,rest) = splitAt s ps
#elif (SKEL==1)
	-- FARM
    xList_all = (farmND np 1 get_homSol primes')

#else
-- # error "no SKEL defined for Eden "
    xList_all = gen_xList primes'
#endif 
# else /* !defined(__EDEN__) */
    {- original sequential version -}
    xList_all = gen_xList primes'
#endif

    -- Input: Infinite list of primes
    -- Output: Infinite list of solutions of a*x=b in Z_p
    gen_xList :: [Integer] -> [[Integer]]
    gen_xList ps = map get_homSol ps

    get_homSol :: Integer -> [Integer]
    get_homSol p    = let 
			b0 = vecHom p b
			a0 = matHom p a
			modDet = modHom p (slow_determinant a0)

			-- homSol = (p : modDet : pmx) -- `using` sPar modDet
#if defined(__EDEN__) && defined(PARVECTOR)
# if (PARVECTOR==1)
			pmx = parMap 
# else 
			pmx = farm PARVECTOR 1
# endif
				(\j -> modHom p (determinant (replaceColumn j a0 b0 ))) [jLo..jHi]
#else                       
			pmx = [ modHom p (determinant (replaceColumn j a0 b0 ))
			      | j <- [jLo..jHi] ]
#endif
			((iLo,jLo),(iHi,jHi)) = matBounds a
		      in
                        -- compute the first two elements; depending on the 
                        -- result force the rest 
# ifdef TRACE
                        trace ("prime " ++ (show p) ++ " has a modDet of " ++ (show modDet) ++ " and a solution vector of " ++ (show (if modDet == 0 then [0]  else pmx)))
# endif
                        (p : modDet : if modDet == 0 then [0] 
						     else pmx)  
			{- the above case is needed for Eden, otherwise
			   the solution vector is also tried to be computed 
			   instead of the GpH version. (UK)
			-}

   {- Step3: -}
   {- uses sequential CRA over lists -}

    -- how many list elems are needed until their product exceeds pBound?
    count :: [Integer] -> Integer
    count = toInteger . (+1) . length . takeWhile (<=pBound) . scanl (*) 1

    noOfPrimes = count (primeList) -- guess how many primes needed
    primeList = projection 0 xList -- primes (the bases for the hom ims)
    detList = projection 1 xList   -- dets in all hom ims

	{- in det and x_i prime replaced by primeList! (UK) -}
    det = snd (seq_list_CRA pBound {-noOfPrimes-} primeList detList detList)

    x_i i = snd (seq_list_CRA pBound {-noOfPrimes-} primeList x_i_List detList)
            where    
	          x_i_List = projection (i+2) xList

    xs = map x_i [0..n-1]         -- overall solution
    x = vector xs

   {- Step4: -}
    gcdVec = vecGcd x
    gcdVal = (gcd gcdVec det) -- `using` sPar det

    -- Solution i.e. a * (res_a/res_b) * res_x = b
    res_a :: Integer
    res_a = gcdVec `div` gcdVal
    res_b :: Integer
    res_b = det `div` gcdVal
    res_x :: Vector Integer
    res_x = vecScalarQuot gcdVec x

#if defined(STRAT) && (defined(__PARALLEL_HASKELL__) || defined(__GRANSIM__))
    {- Aux functions used in the strategies below -}    
    seqListAccum :: Integer -> Strategy [Integer] -> Strategy [[Integer]]
    seqListAccum accum s = \ (xs:xss) -> if accum>pBound
                                            then ()
			                    else s xs `Strategies.seq` 
					         seqListAccum (accum*(head xs)) s xss

    parListAccum :: Integer -> Strategy [Integer] -> Strategy [[Integer]]
    parListAccum accum s = \ (xs:xss) -> if accum>pBound
                                            then ()
			                    else s xs `par` 
					         parListAccum (accum*(head xs)) s xss

    sol_strat :: Strategy [Integer]
    sol_strat = \ (p:modDet:pmx) -> rnf modDet `Strategies.seq` 
                     	      	    if modDet /= 0
                                      then seqList rnf pmx 
                                      else ()

    par_sol_strat :: Strategy [Integer]
    par_sol_strat = \ (p:modDet:pmx) -> rnf modDet `Strategies.seq` 
			                if modDet /= 0
			                  then parList rnf pmx 
			                  else ()

    par_par_sol_strat :: Strategy [Integer]
    par_par_sol_strat = \ (p:modDet:pmx) -> rnf modDet `par` 
			                    if modDet /= 0
			                      then parList rnf pmx 
			                      else ()
#endif
  in 
#ifdef TRACE
	 trace (" pBound: " ++ (show pBound) ++ "; approx. lucky primes " ++ (show guessedNoOfPrimes))
#endif
    (vecScalarQuot gcdVec x, gcdVec `div` gcdVal, det `div` gcdVal)
#if defined(__PARALLEL_HASKELL__) || defined(__GRANSIM__)
    `using`
    \ res -> 
# if (STRAT==0)
           -- (0) seq evaluation
           rwhnf res
# elif (STRAT==1)
           -- (1) (old) straightforward version
           -- This strat forces eval of det before starting eval of hom sols
           rnf noOfPrimes                  `Strategies.seq`
           rnf det                         `Strategies.seq`
           parListN noOfPrimes (\ sol -> let 
                                           modDet = sol!!1
					   pmx = drop 2 sol
					 in 
					   rnf modDet `Strategies.seq` 
			                   if modDet /= 0
			                     then parList rnf pmx
			                     else ())              xList `par`
	   parList rnf xs
# elif (STRAT==2)
           -- (2) no par vect
           -- Avoid parList of result vector otherwise as clean strat
           rnf noOfPrimes                  `Strategies.seq`
           parListN noOfPrimes (\ sol -> let 
                                           modDet = sol!!1
					   pmx = drop 2 sol
					 in 
					   rnf modDet `Strategies.seq` 
			                   if modDet /= 0
			                     then parList rnf pmx
			                     else ())              xList
# elif (STRAT==3)
           -- (3) spec computation
           -- Do not check for modDet in strat; spec eval over hom ims
           rnf noOfPrimes                  `Strategies.seq`
           parListN noOfPrimes (parList rnf) xList `par`
	   parList rnf xs
# elif (STRAT==4)
           -- (4) no forcing of noOfPrimes
           -- Do not force noOfPrimes before using it in parListN
           parListN noOfPrimes (\ sol -> let 
                                           modDet = sol!!1
					   pmx = drop 2 sol
					 in 
					   rnf modDet `Strategies.seq` 
			                   if modDet /= 0
			                     then parList rnf pmx
			                     else ())              xList `par`
					     
	   parList rnf xs
# elif (STRAT==5)
           -- (5) guess number of solutions
           -- Use a hard-wired guess; this doesn't need noOfPrimes at all
           parListN 19 (\ sol -> let 
                                           modDet = sol!!1
					   pmx = drop 2 sol
					 in 
					   rnf modDet `Strategies.seq` 
			                   if modDet /= 0
			                     then parList rnf pmx
			                     else ())              xList `par`
					     
	   parList rnf xs
# elif (STRAT==6)
--           -- (6) seq strat
--           -- Use seq as strat comb; evaluation order complete determined by strat
--           rnf noOfPrimes                  `Strategies.seq`
--           parListN noOfPrimes (\ sol -> let 
--                                           modDet = sol!!1
--					   pmx = drop 2 sol
--					 in 
--					   rnf modDet `Strategies.seq` 
--			                   if modDet /= 0
--			                     then parList rnf pmx
--			                     else ())              xList `Strategies.seq`
					     
--	   parList rnf xs

           -- (6) parListAccum version (old) stfwd version but with 
           -- rnf noOfPrimes                  `Strategies.seq`
	   rnf det                         `par`
	   -- parList rwhnf xs                `Strategies.seq` -- `par` NO!
           parListAccum 1 (\ sol ->      let 
                                           modDet = sol!!1
					   pmx = drop 2 sol
					 in 
					   rnf modDet `Strategies.seq` 
			                   if modDet /= 0
			                     then parList rnf pmx 
			                     else ())              xList `par`
                                                                   --   ^^^^^^
	   parList rwhnf xs
# elif (STRAT==11)
           -- (7) (old) stfwd version (seq, seq)
	   rnf det                         `Strategies.seq`
           seqListAccum 1 sol_strat xList  `par`
	   parList rnf xs
# elif (STRAT==12)
           -- (7) variant of (old) stfwd version (seq,par)
	   rnf det                             `Strategies.seq`
           seqListAccum 1 par_sol_strat xList  `par`
	   parList rnf xs
# elif (STRAT==13)
           -- (7) variant of (old) stfwd version (seq,par) without forcing det
	   -- rnf det                             `Strategies.seq`
           seqListAccum 1 par_sol_strat xList  `par`
	   parList rnf xs
# elif (STRAT==14)
           -- (8) (old) improved version (seqListN)
           rnf noOfPrimes                           `Strategies.seq`
           seqListN noOfPrimes par_sol_strat xList  `par`
	   parList rnf xs
# elif (STRAT==16)
           -- (16) (new, paper) stfwd version
	   rnf det                             `Strategies.seq`
           seqListAccum 1 par_sol_strat xList  `par`
	   parList rnf xs
# elif (STRAT==17)
           -- (17) (new, paper) improved version
	   -- rnf det                                 `Strategies.seq`
           seqListAccum 1 par_par_sol_strat xList  `par`
	   parList rnf xs
# elif (STRAT==18)
           -- (18) (new, paper) final version
           -- Uses inf lists and test on det; the clean strategy
           rnf noOfPrimes                               `Strategies.seq`
           parListN noOfPrimes par_par_sol_strat xList  `par`
	   parList rnf xs
-- 26, 27, 28 are the right strategies! 
-- see ~/ghc-build/test/linsolv/clean-up/all/logs-PhD2/LS-45{.out,*ps.gz}
# elif (STRAT==26)
           -- (26) (new, paper) stfwd version (like 16 but seq_sol_strat)
	   rnf det                             `Strategies.seq`
           seqListAccum 1 sol_strat xList      `par`
	   parList rnf xs
# elif (STRAT==27)
           -- (27) (new, paper) improved version  (like 17 but par_sol_strat)
	   -- rnf det                                 `Strategies.seq`
           seqListAccum 1 par_sol_strat xList      `par`
	   parList rnf xs
# elif (STRAT==28)
           -- (28) (new, paper) final version (like 18 but par_sol_strat)
           -- Uses inf lists and test on det; the clean strategy
	   {- during the evaluation of noOfPrimes, all the determinants in 
	      homomorphic images are evaluated sequentially on top level,
	      only the evaluation of one determinant at a time is parallelised!
	      So this leads to a two stage scheme for the parallelisation.
	      (UK) 
           -}
           rnf guessedNoOfPrimes                               `Strategies.seq`
           parListN guessedNoOfPrimes par_sol_strat xList      `par`
	   parList rnf xs
# else	   
           -- (9) (old) final version
           -- Uses inf lists and test on det; the clean strategy
           rnf noOfPrimes                           `Strategies.seq`
           parListN noOfPrimes par_sol_strat xList  `par`
	   parList rnf xs
# endif /* STRAT */
#endif
