--  Time-stamp: <Wed Sep 29 2010 21:29:32 Stardate: Stardate: [-28]3759.06 hwloidl>
--  $Id: LinSolvS.hs,v 1.1.2.11 2002/07/15 00:33:23 hwloidl Exp $
--
--  This is a `clean' copy of LinSolv.hs, getting rid of debugging stuff
--  and CPPs for Eden etc. Now uses command line arg for choosing parallel
--  version (-DSTRAT is now redundant!).
--  ToDo: Merge this with LinSolv.hs again and nuke this file.
--
--  Module for computing the solution of a system of linear system of equations
--  in several variables.
--  The rhs of the system of equations is the first parameter (given as a 
--  square matrix). The lhs of the system is the second parameter (given as a
--  vector).

--  Note: This version uses Strategies to specify parallelism. It defines the 
--        Strategies on top-level i.e. in the body of the let in the main 
--        linsolv function. Of the various Strategies used, Strategies 26,
--        27 and 28 correspond to those discussed in my LinSolv paper:
--          http://www.cee.hw.ac.uk/~dsg/gph/papers/ps/linsolv.ps.gz
-- 
--  Variants (CPP vars):
--    LU_DET ... use LU decomposition for computing determinant
--    MOD_LU ... use LU decomposition and fwd/bwd substitution, all over
--               a modular image, to compute the solution in the hom im
--               (otherwise Cramer's rule is used)
---------------------------------------------------------------------------- -}

-- @node LinSolv top level function, , ADT Matrix, Top
-- @chapter LinSolv top level function

module LinSolvS(linSolv) where

-- version = "LinSolv: $Revision: 1.1.2.11 $"

-- @node Imports, Auxiliary functions
-- @section Imports

-- @include ModArithm.hs
import ModArithm {- (modHom, Hom(hom))-}
#if defined(STRATEGIES)
# if defined(EVAL_STRATEGIES)
import Control.Parallel
import Control.Parallel.Strategies
import qualified Control.Parallel.SeqStrategies as SeqStrat
# elif defined(SM_STRATEGIES)
import Control.Parallel
import Control.Parallel.Strategies
# else
import Control.Parallel
import Control.Parallel.Strategies
# endif
#endif

-- @include Matrix-list.hs
import Matrix (SqMatrix, Vector, vector, unvector, matBounds, unSqMatrix,
               determinant, 
               vecCont, matCont,
               transp, replaceColumn, size,
               maxElem, maxElemVec, scalarMult, vecScalarQuot,
               matGcd, vecGcd, matHom, vecHom) 

-- import qualified Matrix (determinant)

-- @include CRA.hs
import CRA (seq_list_CRA)

-- @include LUDecomp.hs
import LUDecompMatrix (det, homsolv)  -- det computation via LU decomposition

-- @include Strategies
-- import Strategies

-- @include Ratio.hs
import Ratio

--import Array
{-
slow_determinant :: SqMatrix Integer -> Integer
slow_determinant = Matrix.determinant
-}

determinant :: Integer -> SqMatrix Integer -> Integer
determinant = det


homsolv0 :: Integer -> SqMatrix Integer -> Vector Integer -> [Integer] --Vector Integer
homsolv0 p a0 b0 =
  let
    v  = homsolv p a0 b0
  in
  vecCont v

-- ---------------------------------------------------------------------------
-- @node Auxiliary functions, Main Fct, Imports
-- @section Auxiliary functions
-- ---------------------------------------------------------------------------

primes :: [Integer]
primes = 2 : oddPrimesFrom 3
         where oddPrimesFrom n | isPrime n primes = n : oddPrimesFrom (n+2)
                               | otherwise        =     oddPrimesFrom (n+2)

isPrime :: Integer -> [Integer] -> Bool
isPrime n (l1:ls) | n < l1*l1        = True
                  | n `mod` l1 == 0  = False
                  | otherwise        = isPrime n ls

projection :: Integer -> [[Integer]] -> [Integer]
projection i = {-_scc_ "projection"-} ( map (!! (fromInteger i)) )
 
fact :: Integer -> Integer
fact 0   = 1
fact (n+1) = (n+1) * fact n

-- ---------------------------------------------------------------------------
-- @node Main Fct, Junk, Auxiliary functions
-- @section Main Fct
--
-- Function for solving a system of linear equations
-- ---------------------------------------------------------------------------

-- @cindex linSolv
--  linSolv :: Int ->                             -- version (i.e. STRAT)
--             SqMatrix Integer ->                -- nxn matrix A
--             Vector Integer ->                  -- n vector b
--             (Vector Integer, Integer, Integer) -- n vector x s.t. A*x=b
linSolv v a b = 
  let 
   {- Step1: -}
    n = toInteger (size a) 
    s = max (maxElem a) (maxElemVec b)
    pBound :: Integer
    pBound = 2 * s^n * fact n 

   {- Step2: Computation of solutions in Z/p -}
    primesB = drop 200 primes

    guessedNoOfPrimes :: Integer
    guessedNoOfPrimes =  count primesB
    
    {- original sequential version -}  
    xList = gen_xList primesB             -- xList_all

    -- Input: Infinite list of primes
    -- Output: Infinite list of solutions of a*x=b in Z_p
    gen_xList :: [Integer] -> [[Integer]]
    gen_xList ps = map get_homSol ps

    get_homSol :: Integer -> [Integer]
    get_homSol p = let 
                     b0 = vecHom p b
                     a0 = matHom p a
                     modDet = modHom p (LinSolvS.determinant p a0)
 
			-- homSol = (p : modDet : pmx) -- `using` sPar modDet
                     pmx = homsolv0 p a0 b0
                              
                     ((iLo,jLo),(iHi,jHi)) = matBounds a
		      in
                        -- compute the first two elements; depending on the 
                        -- result force the rest 

                        (p : modDet : if modDet == 0 then [0] 
						     else pmx)  
			{- the above case is needed for Eden, otherwise
			   the solution vector is also tried to be computed 
			   instead of the GpH version. (UK)
			-}

   {- Step3: -}
   {- uses sequential CRA over lists -}

    -- how many list elems are needed until their product exceeds pBound?
    count :: [Integer] -> Integer
    count = toInteger . (+1) . length . takeWhile (<=pBound) . scanl (*) 1

    noOfPrimes = count (primeList) -- guess how many primes needed
    primeList = projection 0 xList -- primes (the bases for the hom ims)
    detList = projection 1 xList   -- dets in all hom ims

	{- in det and x_i prime replaced by primeList! (UK) -}
    det = snd (seq_list_CRA pBound primeList detList detList)

    x_i i = snd (seq_list_CRA pBound primeList x_i_List detList)
            where    
	          x_i_List = projection (i+2) xList

    xs = map x_i [0..n-1]         -- overall solution
    x = vector xs

   {- Step4: -}
    gcdVec = vecGcd x
    gcdVal = (gcd gcdVec det) -- `using` sPar det

    -- Solution i.e. a * (res_a/res_b) * res_x = b
    res_a :: Integer
    res_a = gcdVec `div` gcdVal
    res_b :: Integer
    res_b = det `div` gcdVal
    res_x :: Vector Integer
    res_x = vecScalarQuot gcdVec x

#if defined(STRATEGIES)
# if defined(EVAL_STRATEGIES)
    {- Aux functions used in the strategies below -}    
    seqListAccum :: Integer -> Strategy [Integer] -> Strategy [[Integer]]
    seqListAccum accum s = \ z@(xs:xss) -> if accum>pBound
                                            then r0 z
			                    else do 
                                                   xs' <- s xs
					           xss' <- seqListAccum (accum*(head xs)) s xss
                                                   return (xs':xss')

    parListAccum :: Integer -> Strategy [Integer] -> Strategy [[Integer]]
    parListAccum accum s = \ z@(xs:xss) -> if accum>pBound
                                            then r0 z
			                    else do 
                                                   xs' <- (rpar `dot` s) xs
					           xss' <- parListAccum (accum*(head xs)) s xss
                                                   return (xs':xss')

    sol_strat :: Strategy [Integer]
    sol_strat = \ z@(p:modDet:pmx) -> do 
                                        modDet' <- rseq modDet
                     	      	    	if modDet' /= 0
                                    	  then do 
                                                 pmx' <- (ins (SeqStrat.seqList SeqStrat.rdeepseq)) pmx 
                                                 return (p:modDet':pmx')
                                    	  else r0 z

    par_sol_strat :: Strategy [Integer]
    par_sol_strat = \ z@(p:modDet:pmx) -> do 
                                            modDet' <- rseq modDet
                     	      	    	    if modDet' /= 0
                                    	      then do 
                                                    pmx' <- parList rdeepseq pmx 
                                                    return (p:modDet':pmx')
                                    	      else r0 z

    par_par_sol_strat :: Strategy [Integer]
    par_par_sol_strat = \ z@(p:modDet:pmx) -> do 
                                            	modDet' <- rpar modDet
                     	      	    	    	if modDet' /= 0
                                    	    	  then do 
                                            	        pmx' <- parList rdeepseq pmx 
                                            	        return (p:modDet':pmx')
                                    	    	  else r0 z


    strategy = 
     case v of 
-- 26, 27, 28 are the right strategies! 
-- see ~/ghc-build/test/linsolv/clean-up/all/logs-PhD2/LS-45{.out,*ps.gz}
         26 -> \ res -> 
           -- (26) (new, paper) stfwd version (like 16 but seq_sol_strat)
           do 
	      det' <- rdeepseq det
              z <- (rpar `dot` seqListAccum 1 sol_strat) xList
	      xs' <- parList rdeepseq xs
              return res
         27 -> \ res -> 
           -- (27) (new, paper) improved version  (like 17 but par_sol_strat)
	   -- rnf det                                 `pseq`
           do
              z <- (rpar `dot` seqListAccum 1 par_sol_strat) xList
	      xs' <- parList rdeepseq xs
              return res
         28 -> \ res -> 
           -- (28) (new, paper) final version (like 18 but par_sol_strat)
           -- Uses inf lists and test on det; the clean strategy
	   {- during the evaluation of noOfPrimes, all the determinants in 
	      homomorphic images are evaluated sequentially on top level,
	      only the evaluation of one determinant at a time is parallelised!
	      So this leads to a two stage scheme for the parallelisation.
	      (UK) 
           -}
           do
             z <- rseq guessedNoOfPrimes
             zs <- (rpar `dot` parListN (fromIntegral guessedNoOfPrimes) par_sol_strat) xList
	     xs' <- parList rdeepseq xs
             return res
         29 -> \ res -> 
           -- (29) (21st century version) larger input (LU-det!)
           -- same as 28 but seq strategies because larger inputs we 
           -- use with LU-det generate sufficient top level par already	 
           do 
             z <- rseq guessedNoOfPrimes
             zs <- (rpar `dot` parListN (fromIntegral guessedNoOfPrimes) sol_strat) xList
	     xs' <- parList rdeepseq xs
             return res
         31 -> \ res -> 
           -- (31) with LU-based hom im sol
           -- faster solution means more input; don't need that much par 
           do 
             z <- rseq guessedNoOfPrimes
             zs <- (rpar `dot` parListN (fromIntegral guessedNoOfPrimes) sol_strat) xList
             return res
         32 -> \ res -> 
           -- (32) try to eliminate seq tail in 31
           do 
             z <- rseq guessedNoOfPrimes
             zs <- (rpar `dot` parListN (fromIntegral guessedNoOfPrimes) sol_strat) xList
             xs' <- parList rseq xs
             return res
         _ -> error "unsupported version"   
# elif defined(SM_STRATEGIES)
    par_sol_strat :: Strategy [Integer]
    par_sol_strat = \ z@(p:modDet:pmx) -> do 
                                            modDet' <- rwhnf modDet
                     	      	    	    if modDet' /= 0
                                    	      then do 
                                                    pmx' <- parList rdeepseq pmx 
                                                    return (p:modDet':pmx')
                                    	      else r0 z

    strategy = 
     case v of 
         28 -> \ res -> 
           -- (28) (new, paper) final version (like 18 but par_sol_strat)
           -- Uses inf lists and test on det; the clean strategy
	   {- during the evaluation of noOfPrimes, all the determinants in 
	      homomorphic images are evaluated sequentially on top level,
	      only the evaluation of one determinant at a time is parallelised!
	      So this leads to a two stage scheme for the parallelisation.
	      (UK) 
           -}
           do
             z <- rwhnf guessedNoOfPrimes
             zs0 <- parListN (fromIntegral guessedNoOfPrimes) par_sol_strat xList
             zs0' <- rpar zs0
	     xs' <- parList rdeepseq xs
             return res
         _ -> error "unsupported version"   
# else
-- OLD strategies here!!
    {- Aux functions used in the strategies below -}    
    seqListAccum :: Integer -> Strategy [Integer] -> Strategy [[Integer]]
    seqListAccum accum s = \ z@(xs:xss) -> if accum>pBound
                                            then r0 z
			                    else s xs `pseq` 
					         seqListAccum (accum*(head xs)) s xss

    parListAccum :: Integer -> Strategy [Integer] -> Strategy [[Integer]]
    parListAccum accum s = \ z@(xs:xss) -> if accum>pBound
                                            then r0 z
			                    else s xs `par` 
					         parListAccum (accum*(head xs)) s xss

    sol_strat :: Strategy [Integer]
    sol_strat = \ z@(p:modDet:pmx) -> rnf modDet `pseq` 
                     	      	    if modDet /= 0
                                      then seqList rnf pmx 
                                      else r0 z -- return z

    par_sol_strat :: Strategy [Integer]
    par_sol_strat = \ z@(p:modDet:pmx) -> rnf modDet `pseq` 
			                if modDet /= 0
			                  then parList rnf pmx 
			                  else r0 z -- return z

    par_par_sol_strat :: Strategy [Integer]
    par_par_sol_strat = \ z@(p:modDet:pmx) -> rnf modDet `par` 
			                      if modDet /= 0
			                      then parList rnf pmx 
			                      else r0 z -- return z


    strategy = 
     case v of 
-- 26, 27, 28 are the right strategies! 
-- see ~/ghc-build/test/linsolv/clean-up/all/logs-PhD2/LS-45{.out,*ps.gz}
         26 -> \ res -> 
           -- (26) (new, paper) stfwd version (like 16 but seq_sol_strat)
	   rnf det                             `pseq`
           seqListAccum 1 sol_strat xList      `par`
	   parList rnf xs
			                   `pseq` r0 res {- return res -}
         27 -> \ res -> 
           -- (27) (new, paper) improved version  (like 17 but par_sol_strat)
	   -- rnf det                                 `pseq`
           seqListAccum 1 par_sol_strat xList      `par`
	   parList rnf xs
			                   `pseq` r0 res {- return res -}
         28 -> \ res -> 
           -- (28) (new, paper) final version (like 18 but par_sol_strat)
           -- Uses inf lists and test on det; the clean strategy
	   {- during the evaluation of noOfPrimes, all the determinants in 
	      homomorphic images are evaluated sequentially on top level,
	      only the evaluation of one determinant at a time is parallelised!
	      So this leads to a two stage scheme for the parallelisation.
	      (UK) 
           -}
           rnf guessedNoOfPrimes                               `pseq`
           parListN (fromIntegral guessedNoOfPrimes) par_sol_strat xList      `par`
	   parList rnf xs
			                   `pseq` r0 res {- return res -}
         29 -> \ res -> 
           -- (29) (21st century version) larger input (LU-det!)
           -- same as 28 but seq strategies because larger inputs we 
           -- use with LU-det generate sufficient top level par already	 
           rnf guessedNoOfPrimes                               `pseq`
           parListN (fromIntegral guessedNoOfPrimes) sol_strat xList      `par`
	   parList rnf xs
			                   `pseq` r0 res {- return res -}
         31 -> \ res -> 
           -- (31) with LU-based hom im sol
           -- faster solution means more input; don't need that much par 
           rnf guessedNoOfPrimes                               `pseq`
           parListN (fromIntegral guessedNoOfPrimes) sol_strat xList
			                   `pseq` r0 res {- return res -}
         32 -> \ res -> 
           -- (32) try to eliminate seq tail in 31
           rnf guessedNoOfPrimes                               `pseq`
           parListN (fromIntegral guessedNoOfPrimes) sol_strat xList          `par`
           parList rwhnf xs
			                   `pseq` r0 res {- return res -}
#endif /* EVAL_STRATEGIES */
#endif /* STRATEGIES */
  in 
    (vecScalarQuot gcdVec x, gcdVec `div` gcdVal, det `div` gcdVal,
     guessedNoOfPrimes, 
     count $ map snd $ (filter (\ (d,p) -> d/=0 )) $ zip detList primeList )
#if defined(STRATEGIES)
    `using` strategy
#endif


-- the full range of strategies used in the past; not ported to new strats yet
#if 0
#if defined(STRATEGIES)
# if defined(EVAL_STRATEGIES)
    {- Aux functions used in the strategies below -}    
    seqListAccum :: Integer -> Strategy [Integer] -> Strategy [[Integer]]
    seqListAccum accum s = \ z@(xs:xss) -> if accum>pBound
                                            then r0 z
			                    else s xs `pseq` 
					         seqListAccum (accum*(head xs)) s xss

    parListAccum :: Integer -> Strategy [Integer] -> Strategy [[Integer]]
    parListAccum accum s = \ z@(xs:xss) -> if accum>pBound
                                            then r0 z
			                    else s xs `par` 
					         parListAccum (accum*(head xs)) s xss

    sol_strat :: Strategy [Integer]
    sol_strat = \ z@(p:modDet:pmx) -> rnf modDet `pseq` 
                     	      	    if modDet /= 0
                                      then seqList rnf pmx 
                                      else r0 z -- return z

    par_sol_strat :: Strategy [Integer]
    par_sol_strat = \ z@(p:modDet:pmx) -> rnf modDet `pseq` 
			                if modDet /= 0
			                  then parList rnf pmx 
			                  else r0 z -- return z

    par_par_sol_strat :: Strategy [Integer]
    par_par_sol_strat = \ z@(p:modDet:pmx) -> rnf modDet `par` 
			                      if modDet /= 0
			                      then parList rnf pmx 
			                      else r0 z -- return z


    strategy = 
     case v of 
         0 -> \ res -> 
           -- (0) seq evaluation
           rwhnf res
         1 -> \ res -> 
           -- (1) (old) straightforward version
           -- This strat forces eval of det before starting eval of hom sols
           rnf noOfPrimes                  `pseq`
           rnf det                         `pseq`
           parListN (fromIntegral noOfPrimes)
                               (\ sol -> let 
                                           modDet = sol!!1
					   pmx = drop 2 sol
					 in 
					   rnf modDet `pseq` 
			                   if modDet /= 0
			                     then parList rnf pmx
			                     else r0 sol {- return sol -})              xList `par`
	   parList rnf xs
			                   `pseq` r0 res {- return res -}

         2 -> \ res -> 
           -- (2) no par vect
           -- Avoid parList of result vector otherwise as clean strat
           rnf noOfPrimes                  `pseq`
           parListN (fromIntegral noOfPrimes)
                               (\ sol -> let 
                                           modDet = sol!!1
					   pmx = drop 2 sol
					 in 
					   rnf modDet `pseq` 
			                   if modDet /= 0
			                     then parList rnf pmx
			                     else r0 sol {- return sol -})              xList
			                   `pseq` r0 res {- return res -}


         3 -> \ res -> 
           -- (3) spec computation
           -- Do not check for modDet in strat; spec eval over hom ims
           rnf noOfPrimes                  `pseq`
           parListN (fromIntegral noOfPrimes) (parList rnf) xList `par`
	   parList rnf xs
			                   `pseq` r0 res {- return res -}
         4 -> \ res -> 
           -- (4) no forcing of noOfPrimes
           -- Do not force noOfPrimes before using it in parListN
           parListN (fromIntegral noOfPrimes) (\ sol -> let 
                                           modDet = sol!!1
					   pmx = drop 2 sol
					 in 
					   rnf modDet `pseq` 
			                   if modDet /= 0
			                     then parList rnf pmx
			                     else r0 sol {- return sol -})              xList `par`
	   parList rnf xs
			                   `pseq` r0 res {- return res -}

         5 -> \ res -> 
           -- (5) guess number of solutions
           -- Use a hard-wired guess; this doesn't need noOfPrimes at all
           parListN 19 (\ sol -> let 
                                           modDet = sol!!1
					   pmx = drop 2 sol
					 in 
					   rnf modDet `pseq` 
			                   if modDet /= 0
			                     then parList rnf pmx
			                     else r0 sol {- return sol -})              xList `par`
	   parList rnf xs
			                   `pseq` r0 res {- return res -}
         6 -> \ res -> 
--           -- (6) seq strat
--           -- Use seq as strat comb; evaluation order complete determined by strat
--           rnf noOfPrimes                  `pseq`
--           parListN (fromIntegral noOfPrimes) (\ sol -> let 
--                                           modDet = sol!!1
--					   pmx = drop 2 sol
--					 in 
--					   rnf modDet `pseq` 
--			                   if modDet /= 0
--			                     then parList rnf pmx
--			                     else ())              xList `pseq`
					     
--	   parList rnf xs

           -- (6) parListAccum version (old) stfwd version but with 
           -- rnf noOfPrimes                  `pseq`
	   rnf det                         `par`
	   -- parList rwhnf xs                `pseq` -- `par` NO!
           parListAccum 1 (\ sol ->      let 
                                           modDet = sol!!1
					   pmx = drop 2 sol
					 in 
					   rnf modDet `pseq` 
			                   if modDet /= 0
			                     then parList rnf pmx 
			                     else r0 sol {- return sol -})              xList `par`
                                                                   --   ^^^^^^
	   parList rwhnf xs
			                   `pseq` r0 res {- return res -}
         11 -> \ res -> 
           -- (7) (old) stfwd version (seq, seq)
	   rnf det                         `pseq`
           seqListAccum 1 sol_strat xList  `par`
	   parList rnf xs
			                   `pseq` r0 res {- return res -}
         12 -> \ res -> 
           -- (7) variant of (old) stfwd version (seq,par)
	   rnf det                             `pseq`
           seqListAccum 1 par_sol_strat xList  `par`
	   parList rnf xs
			                   `pseq` r0 res {- return res -}
         13 -> \ res -> 
           -- (7) variant of (old) stfwd version (seq,par) without forcing det
	   -- rnf det                             `pseq`
           seqListAccum 1 par_sol_strat xList  `par`
	   parList rnf xs
			                   `pseq` r0 res {- return res -}
         14 -> \ res -> 
           -- (8) (old) improved version (seqListN)
           rnf noOfPrimes                           `pseq`
           seqListN (fromIntegral noOfPrimes) par_sol_strat xList  `par`
	   parList rnf xs
			                   `pseq` r0 res {- return res -}
         16 -> \ res -> 
           -- (16) (new, paper) stfwd version
	   rnf det                             `pseq`
           seqListAccum 1 par_sol_strat xList  `par`
	   parList rnf xs
			                   `pseq` r0 res {- return res -}
         17 -> \ res -> 
           -- (17) (new, paper) improved version
	   -- rnf det                                 `pseq`
           seqListAccum 1 par_par_sol_strat xList  `par`
	   parList rnf xs
			                   `pseq` r0 res {- return res -}
         18 -> \ res -> 
           -- (18) (new, paper) final version
           -- Uses inf lists and test on det; the clean strategy
           rnf noOfPrimes                               `pseq`
           parListN (fromIntegral noOfPrimes) par_par_sol_strat xList  `par`
	   parList rnf xs
			                   `pseq` r0 res {- return res -}

-- 26, 27, 28 are the right strategies! 
-- see ~/ghc-build/test/linsolv/clean-up/all/logs-PhD2/LS-45{.out,*ps.gz}
         26 -> \ res -> 
           -- (26) (new, paper) stfwd version (like 16 but seq_sol_strat)
	   rnf det                             `pseq`
           seqListAccum 1 sol_strat xList      `par`
	   parList rnf xs
			                   `pseq` r0 res {- return res -}
         27 -> \ res -> 
           -- (27) (new, paper) improved version  (like 17 but par_sol_strat)
	   -- rnf det                                 `pseq`
           seqListAccum 1 par_sol_strat xList      `par`
	   parList rnf xs
			                   `pseq` r0 res {- return res -}
         28 -> \ res -> 
           -- (28) (new, paper) final version (like 18 but par_sol_strat)
           -- Uses inf lists and test on det; the clean strategy
	   {- during the evaluation of noOfPrimes, all the determinants in 
	      homomorphic images are evaluated sequentially on top level,
	      only the evaluation of one determinant at a time is parallelised!
	      So this leads to a two stage scheme for the parallelisation.
	      (UK) 
           -}
           rnf guessedNoOfPrimes                               `pseq`
           parListN (fromIntegral guessedNoOfPrimes) par_sol_strat xList      `par`
	   parList rnf xs
			                   `pseq` r0 res {- return res -}
         29 -> \ res -> 
           -- (29) (21st century version) larger input (LU-det!)
           -- same as 28 but seq strategies because larger inputs we 
           -- use with LU-det generate sufficient top level par already	 
           rnf guessedNoOfPrimes                               `pseq`
           parListN (fromIntegral guessedNoOfPrimes) sol_strat xList      `par`
	   parList rnf xs
			                   `pseq` r0 res {- return res -}
         31 -> \ res -> 
           -- (31) with LU-based hom im sol
           -- faster solution means more input; don't need that much par 
           rnf guessedNoOfPrimes                               `pseq`
           parListN (fromIntegral guessedNoOfPrimes) sol_strat xList
			                   `pseq` r0 res {- return res -}
         32 -> \ res -> 
           -- (32) try to eliminate seq tail in 31
           rnf guessedNoOfPrimes                               `pseq`
           parListN (fromIntegral guessedNoOfPrimes) sol_strat xList          `par`
           parList rwhnf xs
			                   `pseq` r0 res {- return res -}
         _ ->  \ res -> 
           -- (9) (old) final version
           -- Uses inf lists and test on det; the clean strategy
           rnf noOfPrimes                           `pseq`
           parListN (fromIntegral noOfPrimes) sol_strat xList  `par`
	   parList rnf xs
			                   `pseq` r0 res {- return res -}
#endif /* EVAL_STRATEGIES */
#endif /* STRATEGIES */
#endif 