module TestAux where

import System(getArgs)
import Time
import Random
import Ratio
-- @include LUDecomp.hs
#if defined(LU_MAT)
import LUDecompMatrix   -- det computation via LU decomposition
#else
import LUDecomp 
#endif
       (det, homsolv, mult_mat_vec)
import Matrix 
import Array

---------------------------------------------------------------------------

listToListList :: Int -> [a] -> [[a]]
listToListList c m 
 | length m <= c = [m]
 | otherwise = c1 : listToListList c resto
  where (c1,resto) = splitAt c m

lolArr bds@((xlo,ylo), (xhi, yhi)) zss = 
  array bds (concat [ [ ((x,y), (zss!!(x-1))!!(y-1)) | y <- [ylo..yhi] ] 
                    | x <- [xlo..xhi] ])

---------------------------------------------------------------------------

showTimeDiff (TimeDiff tdYear tdMonth tdDay tdHour tdMin tdSec tdPicosec) 
 | tdYear>0 || tdMonth>0 || tdDay>0 = error "showTimeDiff: time diff too large"
 | otherwise = let
                 tdHour' :: Integer
                 tdHour' = fromIntegral tdHour
                 tdMin' :: Integer
                 tdMin' = fromIntegral tdMin
                 tdSec' :: Integer
                 tdSec' = fromIntegral tdSec
                 ms :: Integer
                 ms = 1000*(60*60*tdHour'+60*tdMin'+tdSec')+(tdPicosec `div` (10^9))
               in
                 show ms

{-
mkMat :: Int -> Integer -> IO (SqMatrix Integer)
mkMat n m
  = do 
      x <- getStdRandom (randomR (1000,2000))
      let
         g = mkStdGen x -- deterministic input: fix x
         l = map (`mod` m) $ randoms g :: [Integer]
         a = lolSqMatrix ((1,1),(n,n)) (listToListList n (take (n*n) l))
      return a
-}

#if !defined(MOD_LU)
mkArrMat :: -- (Fractional a, Random a, Integral a) =>
            Int -> Integer -> IO (Array (Int, Int) (Rational))
mkArrMat n m 
  = do 
      x <- getStdRandom (randomR (1000,2000))
      let
         g = mkStdGen x -- deterministic input: fix x
         l = map (`mod` m) $ randoms g :: [Integer]
         (nums, denoms) = splitAt (n*n) l
         a = lolArr ((1,1),(n,n)) 
               (listToListList n (zipWith (\ x y -> (min x y % max x y)) nums denoms))
      return a
#else
# if defined(LU_MAT)
mkArrMat :: (Integral a, Random a) => 
            Int -> a -> IO (SqMatrix a)
#else
mkArrMat :: (Integral a, Random a) => 
            Int -> a -> IO (Array (Int, Int) a)
# endif /* LU_MAT */
mkArrMat n m 
  = do 
      x <- getStdRandom (randomR (1000,2000))
      let
         g = mkStdGen x -- deterministic input: fix x
         l = map (`mod` m) $ randoms g  -- :: [Integer]
# if defined(LU_MAT)
         a = lolSqMatrix ((1,1),(n,n)) (listToListList n (take (n*n) l))
#else
         a = lolArr ((1,1),(n,n)) (listToListList n (take (n*n) l))
# endif /* LU_MAT */
      return a
#endif

#if !defined(MOD_LU)
mkArrVec :: -- (Fractional a, Random a, Integral a, 
            Int -> Integer -> IO (Array (Int) (Rational))
mkArrVec n m 
  = do 
      x <- getStdRandom (randomR (1000,2000))
      let
         g = mkStdGen x -- deterministic input: fix x
         l = map (`mod` m) $ randoms g  :: [Integer]
         (nums, denoms) = splitAt n l
         a = array (1,n) 
              (zip [1..n] (zipWith (\ x y -> (min x y % max x y)) nums denoms))
      return a
#else
# if defined(LU_MAT)
mkArrVec :: (Integral a, Random a) => 
            Int -> a -> IO (Vector a)
# else
mkArrVec :: (Integral a, Random a) => 
            Int -> a -> IO (Array (Int) a)
# endif /* LU_MAT */
mkArrVec n m 
  = do 
      x <- getStdRandom (randomR (1000,2000))
      let
         g = mkStdGen x -- deterministic input: fix x
         l = map (`mod` m) $ randoms g -- :: [Int]
# if defined(LU_MAT)
         a = vector (take n l)
# else
         a = array (1,n) (zip [1..n] (take n l))
# endif /* LU_MAT */
      return a
#endif

-- The Matrix versions

mkSqMat :: (Integral a, Random a) => 
            Int -> a -> IO (SqMatrix a)
mkSqMat n m 
  = do 
      x <- getStdRandom (randomR (1000,2000))
      let
         g = mkStdGen x -- deterministic input: fix x
         l = map (`mod` m) $ randoms g  -- :: [Integer]
         a = lolSqMatrix ((1,1),(n,n)) (listToListList n (take (n*n) l))
      return a

mkVec :: (Integral a, Random a) => 
            Int -> a -> IO (Vector a)
mkVec n m 
  = do 
      x <- getStdRandom (randomR (1000,2000))
      let
         g = mkStdGen x -- deterministic input: fix x
         l = map (`mod` m) $ randoms g -- :: [Int]
         a = vector (take n l)
      return a

-- Uses the format needed for calc.el
#if !defined(MOD_LU)
showArrMat :: Array (Int, Int) (Ratio Integer) -> String
showArrMat m =  "[" ++
             foldl (\ s c -> if null s then c else s++","++c) [] 
             [ "[" ++ 
               foldl (\ s c -> if null s then c else s++","++c) [] 
                     [ show (numerator x) ++ ":" ++ show (denominator x) 
                     | j <- [1..n], let x = m!(i,j) ]
               ++ "]\n"
               | i <- [1..n] ]  
             ++ "]"
             where 
                bds@((rl,cl),(rh,ch)) = bounds m
                n | rh==ch = rh
                  | otherwise = error ("printMat: non-square; bounds: " ++ (show bds))
#else
showArrMat m =  "[" ++
             foldl (\ s c -> if null s then c else s++","++c) [] 
             [ "[" ++ 
               foldl (\ s c -> if null s then c else s++","++c) [] 
                     [ show x | j <- [1..n], let x = m!(i,j) ]
               ++ "]\n"
               | i <- [1..n] ]  
             ++ "]"
             where 
                bds@((rl,cl),(rh,ch)) = bounds m
                n | rh==ch = rh
                  | otherwise = error ("printMat: non-square; bounds: " ++ (show bds))

showArrVec v = "[" ++ 
               foldl (\ s c -> if null s then c else s++","++c) [] 
                     [ show x | j <- [1..n], let x = v!j ]
               ++ "]\n"
             where 
                bds@(l,h) = bounds v
                n | l==1 = h
                  | otherwise = error ("showArrVec: vector bds don't start at 1' bounds: " ++ (show bds))
#endif

-----------------------------------------------------------------------------

#if !defined(MOD_LU)
# error "NOT MOD_LU STEUP:"
mkInput :: -- (Fractional a, Random a, Integral a,
            Int -> Integer -> IO (Array (Int, Int) (Rational) , Array (Int) (Rational))
mkInput n m =
  do
        a' <- mkArrMat n m
              {-
              if n/=0 then mkArrMat n m -- make that random matrix, except
                      else case m of    -- if n==0, then m is the id of input
                            -- 1 -> return a1    
                            2 -> return a11    
                            3 -> return a12    
                            12 -> return a_dense_12
                            14 -> return a_sparse_14
                            18 -> return a_sparse_18
                            _ -> error "main: invalid matrix version"
               -}
        b' <- mkArrVec n m
        return (a',b')
#else /* MOD_LU */

# if defined(LU_MAT)
mkInput :: (Integral a, Random a) =>
           Int -> a -> IO (SqMatrix a , Vector  a)
# else
mkInput :: (Integral a, Random a) =>
            Int -> a -> IO (Array (Int, Int) a , Array (Int) a)
# endif
mkInput n m =
  do
        a' <- if n/=0 then mkArrMat n m 
                      else error "invalid matrix size 0"
        b' <- mkArrVec n m
        return (a',b')
#endif

#if defined(MOD_LU)

# if defined(LU_MAT)

test_det :: (Integral a) => 
            a -> SqMatrix a -> IO ()
test_det m a' =
  do
# ifdef OUTPUT
        putStrLn ("ArrMatrix a':" ++ show a')
# endif
        putStrLn ("Testing (and timing) det ...")
        c3 <- getClockTime
        let d' = det m a'
# ifdef OUTPUT
        putStrLn ("det a':" ++ show d')
# else
        putStrLn ("signum (det a') = " ++ (show (signum d')))
# endif
        c4 <- getClockTime
        putStrLn ("Total Runtime (in ms): @ " ++ (showTimeDiff (diffClockTimes c4 c3)) ++ " @")

# else

test_det :: (Integral a) => 
            a -> Array (Int, Int) a -> IO ()
test_det m a' =
  do
# ifdef OUTPUT
        putStrLn ("ArrMatrix a':" ++ showArrMat a')
# endif
        putStrLn ("Testing (and timing) det ...")
        c3 <- getClockTime
        let d' = det m a'
# ifdef OUTPUT
        putStrLn ("det a':" ++ show d')
# else
        putStrLn ("signum (det a') = " ++ (show (signum d')))
# endif
        c4 <- getClockTime
        putStrLn ("Total Runtime (in ms): @ " ++ (showTimeDiff (diffClockTimes c4 c3)) ++ " @")

# endif /* LU_MAT */

#else /* !MOD_LU */

test_det a' =
  do
# ifdef OUTPUT
        putStrLn ("ArrMatrix a':" ++ showArrMat a')
# endif
        putStrLn ("Testing (and timing) det ...")
        c3 <- getClockTime
        let d' = det a'
# ifdef OUTPUT
        putStrLn ("det a':" ++ show d')
# else
        putStrLn ("signum (det a') = " ++ (show (signum d')))
# endif
        c4 <- getClockTime
        putStrLn ("Total Runtime (in ms): @ " ++ (showTimeDiff (diffClockTimes c4 c3)) ++ " @")

#endif

#if defined(MOD_LU)

# if defined(LU_MAT)

test_linsolv_lu m a' b' = 
  do
# ifdef OUTPUT
        putStrLn ("Modulus m: " ++ (show m))
        putStrLn ("ArrMatrix a': " ++ show a')
        putStrLn ("ArrVector b': " ++ show b')
# endif
        putStrLn ("Testing (and timing) homsolv a' b' ...")
        c5 <- getClockTime
        let x' = homsolv m a' b'
        let z' = mult_mat_vec m a' x'
# ifdef OUTPUT
        putStrLn ("a' * x' :     " ++ show z')
        putStrLn ("homsolv a' b':" ++ show x')
# else
# if !defined(LU_MAT)
        putStrLn ("sum (map signum x') = " ++ (show (sum (map signum (elems x')))))
# endif
# endif
        c6 <- getClockTime
        putStrLn ("Total Runtime (in ms): @ " ++ (showTimeDiff (diffClockTimes c6 c5)) ++ " @")
        if z'/=b'
          then error $ "ERROR in checking result." -- : "\nb'=" ++ (show b') ++ "\nResult=" ++ (show (mult_mat_vec a' x'))
          else putStrLn "Result OK"         

# else

test_linsolv_lu m a' b' = 
  do
# ifdef OUTPUT
        putStrLn ("Modulus m: " ++ (show m))
        putStrLn ("ArrMatrix a': " ++ showArrMat a')
        putStrLn ("ArrVector b': " ++ showArrVec b')
# endif
        putStrLn ("Testing (and timing) homsolv a' b' ...")
        c5 <- getClockTime
        let x' = homsolv m a' b'
        let z' = mult_mat_vec m a' x'
# ifdef OUTPUT
        putStrLn ("a' * x' :     " ++ showArrVec z')
        putStrLn ("homsolv a' b':" ++ showArrVec x')
# else
        putStrLn ("sum (map signum x') = " ++ (show (sum (map signum (elems x')))))
# endif
        c6 <- getClockTime
        putStrLn ("Total Runtime (in ms): @ " ++ (showTimeDiff (diffClockTimes c6 c5)) ++ " @")
        if z'/=b'
          then error $ "ERROR in checking result." -- : "\nb'=" ++ (show b') ++ "\nResult=" ++ (show (mult_mat_vec a' x'))
          else putStrLn "Result OK"         

# endif /* LU_MAT */

#else /* !MOD_LU ---------------------------------------------------------- */

test_linsolv_lu a' b' = 
  do
# ifdef OUTPUT
        putStrLn ("ArrMatrix a':" ++ showArrMat a')
# endif
        putStrLn ("Testing (and timing) linsolv_lu a' b' ...")
        c5 <- getClockTime
        let x' = linsolv_lu a' b'
# ifdef OUTPUT
        putStrLn ("linsolv_lu a' b':" ++ show x')
# else
        putStrLn ("sum (map signum x') = " ++ (show (sum (map signum (elems x')))))
# endif
        c6 <- getClockTime
        putStrLn ("Total Runtime (in ms): @ " ++ (showTimeDiff (diffClockTimes c6 c5)) ++ " @")
        if (mult_mat_vec a' x')/=b'
          then error $ "ERROR in checking result." -- : "\nb'=" ++ (show b') ++ "\nResult=" ++ (show (mult_mat_vec a' x'))
          else putStrLn "Result OK"         

#endif /* MOD_LU */

--test_it :: Int -> Int -> Int -> IO ()
test_it 0 _ _ = return ()
test_it x n m = 
  do
#if defined(MOD_LU)
        putStrLn ("Testing homsolv with modular arithmetic.")
#else
        putStrLn ("Testing linsolv with rational arithmetic.")
#endif
        putStrLn ((show x) ++ " more iterations to go ...")
        (a',b') <- mkInput n (toInteger m)
        -- Time array-based version
#if defined(MOD_LU)
        test_det m a'
#else
        test_det a'
#endif
        -- Time array-based version
#if defined(MOD_LU)
        test_linsolv_lu m a' b'
#else
        test_linsolv_lu a' b'
#endif
        -- rec call
        test_it (x-1) n m

---------------------------------------------------------------------------

