module BinTree where

import List

-- data types

data BinTree a = Leaf a
               | Node (BinTree a) (BinTree a)
               deriving (Eq) -- (Eq,Read,Show)

mkTree :: [Int] -> BinTree Int
mkTree [x] = Leaf x
mkTree xs  = Node left right
             	 where (l,r) = halve xs
             	       left = mkTree l
             	       right = mkTree r

	     	       halve :: [a] -> ([a],[a])
	     	       halve xs = (take (n `div` 2) xs, drop (n `div` 2) xs)
		                  where n = length xs

flatten :: BinTree a -> [a]
flatten (Node l r)   = (flatten l)++(flatten r)
flatten (Leaf x)     = [x]

-- linear version (see Bird84)
flattenAcc :: BinTree a -> [a] -> [a]
flattenAcc (Leaf x)   ys  = x:ys
flattenAcc (Node l r) ys  = flattenAcc l (flattenAcc r ys)

size :: BinTree a -> Int
size (Leaf _) = 1
size (Node l r) = size l + size r

instance (Show a) => Show (BinTree a) where
 show = show' 0 
        where
       	      show' n (Leaf x) = (take n $ repeat ' ') ++ (show x)
       	      show' n (Node l r) = List.concat (List.intersperse "\n" {- unlines -} [show' (n+1) l, if n==0 then "@" else (take (n-1) $ repeat ' ')++"+", show' (n+1) r])


